(* TYPES POUR L'ARBRE DE SYNTAXE ABSTRAITE *)

type exp_int =
  | Val_r1 (* valeur du registre 1 *)
  | Val_r2 (* valeur du registre 2 *)
  | Val of int (* entier *)
  | Sum_op of exp_int * exp_int (* somme des deux arguments *)
  | Prod_op of exp_int * exp_int (* produit des deux arguments *)

type exp_bool =
  | And_op of exp_bool * exp_bool (* argument 1 ET argument 2 *)
  | Or_op of exp_bool * exp_bool (* argument 1 OU argument 2 *)
  | Equal_op of exp_int * exp_int (* egalite entre les arguments *)
  | Inf_op of exp_int * exp_int (* vrai si argument 1 < argument 2 *)
  | Val_true (* vrai *)
  | Val_false (* faux *)

type ast = 
  | While_loop of exp_bool * ast (* boucle "while", argument 1 : condition, argument 2 : corps *)
  | If_cond of exp_bool * ast * ast (* conditionnelle, argument 1 : condition, argument 2 : alors, argument 3 : sinon *)
  | Affect_r1 of exp_int (* affectation dans le registre 1 *)
  | Affect_r2 of exp_int (* affectation dans le registre 2 *)
  | Seq of ast * ast (* sequence d'instructions *)

(* CALCUL DES VALEURS DES REGISTRES CORRESPONDANT A UN ARBRE DE SYNTAXE ABSTRAITE *)

let rec eval_exp_int exp r1 r2 =
  match exp with
    | Val_r1 -> r1
    | Val_r2 -> r2
    | Val(n) -> n
    | Sum_op(exp1,exp2) -> 
	let v1 = eval_exp_int exp1 r1 r2
	and v2 = eval_exp_int exp2 r1 r2
	in v1 + v2
    | Prod_op(exp1,exp2) ->
	let v1 = eval_exp_int exp1 r1 r2
	and v2 = eval_exp_int exp2 r1 r2
	in v1 * v2	  

let rec eval_exp_bool exp r1 r2 =
  match exp with
    | And_op(exp1,exp2) ->
	let v1 = eval_exp_bool exp1 r1 r2
	and v2 = eval_exp_bool exp2 r1 r2
	in v1 && v2
    | Or_op(exp1,exp2) ->
	let v1 = eval_exp_bool exp1 r1 r2
	and v2 = eval_exp_bool exp2 r1 r2
	in v1 || v2
    | Equal_op(exp1,exp2) ->
	let v1 = eval_exp_int exp1 r1 r2
	and v2 = eval_exp_int exp2 r1 r2
	in v1 = v2
    | Inf_op(exp1,exp2) ->
	let v1 = eval_exp_int exp1 r1 r2
	and v2 = eval_exp_int exp2 r1 r2
	in v1 < v2
    | Val_true -> true
    | Val_false -> false

let rec eval_ast ast r1 r2 =
  match ast with
    | While_loop(exp_b,ast) ->
	let cond = eval_exp_bool exp_b r1 r2
	in 
	  if cond then 
	    let reg1,reg2 = eval_ast ast r1 r2
	    in eval_ast (While_loop (exp_b,ast)) reg1 reg2
	  else
	    r1,r2
    | If_cond(exp_b,ast1,ast2) ->
	let cond = eval_exp_bool exp_b r1 r2
	in
	  if cond then eval_ast ast1 r1 r2
	  else eval_ast ast2 r1 r2
    | Affect_r1(exp_i) ->
	let v = eval_exp_int exp_i r1 r2
	in v,r2
    | Affect_r2(exp_i) ->
	let v = eval_exp_int exp_i r1 r2
	in r1,v
    | Seq(ast1,ast2) ->
	let reg1,reg2 = eval_ast ast1 r1 r2
	in eval_ast ast2 reg1 reg2
